# AUTOGENERATED! DO NOT EDIT! File to edit: 03_tab.export.ipynb (unless otherwise specified).

__all__ = ['load_pandas']

# Cell
from fastai.tabular.all import *

# Cell
@patch
def export(self:TabularPandas, fname='export.pkl', pickle_protocol=2):
    "Export the contents of `self` without the items"
    old_to = self
    self = self.new_empty()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        pickle.dump(self, open(Path(fname), 'wb'), protocol=pickle_protocol)
        self = old_to

# Cell
def load_pandas(fname):
    "Load in a `TabularPandas` object from `fname`"
    distrib_barrier()
    res = pickle.load(open(fname, 'rb'))
    return res


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"get_version": "00_utils.ipynb",
         "state_versions": "00_utils.ipynb",
         "create_timm_body": "02_vision.external.timm.ipynb",
         "create_timm_model": "02_vision.external.timm.ipynb",
         "timm_learner": "02_vision.external.timm.ipynb",
         "TabularPandas.export": "03_tab.export.ipynb",
         "load_pandas": "03_tab.export.ipynb",
         "Categorify": "03_tab.stats.ipynb",
         "Normalize.__init__": "03_tab.stats.ipynb",
         "setups": "03_tab.stats.ipynb",
         "FillMissing": "03_tab.stats.ipynb",
         "Recorder.plot_lr_find": "04_lr_finder.ipynb",
         "Learner.lr_find": "04_lr_finder.ipynb"}

modules = ["utils.py",
           "vision/timm.py",
           "tab/export.py",
           "tab/stats.py",
           "callback/lr_find.py"]

doc_url = "https://walkwithfastai.com/"

git_url = "https://github.com/walkwithfastai/walkwithfastai.github.io/tree/master/"

def custom_doc_links(name):
    from nbdev.showdoc import try_external_doc_link
    return try_external_doc_link(name, ['fastcore', 'nbdev', 'fastai', 'timm'])



# AUTOGENERATED! DO NOT EDIT! File to edit: 02_vision.external.timm.ipynb (unless otherwise specified).

__all__ = ['create_timm_body', 'create_timm_model', 'timm_learner']

# Cell
from fastai.vision.all import *

# Cell
from timm import create_model
from fastai.vision.learner import _update_first_layer

# Cell
def create_timm_body(arch:str, pretrained=True, cut=None, n_in=3):
    "Creates a body from any model in the `timm` library."
    model = create_model(arch, pretrained=pretrained, num_classes=0, global_pool='')
    _update_first_layer(model, n_in, pretrained)
    if cut is None:
        ll = list(enumerate(model.children()))
        cut = next(i for i,o in reversed(ll) if has_pool_type(o))
    if isinstance(cut, int): return nn.Sequential(*list(model.children())[:cut])
    elif callable(cut): return cut(model)
    else: raise NamedError("cut must be either integer or function")

# Cell
def create_timm_model(arch:str, n_out, cut=None, pretrained=True, n_in=3, init=nn.init.kaiming_normal_, custom_head=None,
                     concat_pool=True, **kwargs):
    "Create custom architecture using `arch`, `n_in` and `n_out` from the `timm` library"
    body = create_timm_body(arch, pretrained, None, n_in)
    if custom_head is None:
        nf = num_features_model(nn.Sequential(*body.children())) * (2 if concat_pool else 1)
        head = create_head(nf, n_out, concat_pool=concat_pool, **kwargs)
    else: head = custom_head
    model = nn.Sequential(body, head)
    if init is not None: apply_init(model[1], init)
    return model

# Cell
from fastai.vision.learner import _add_norm

# Cell
def timm_learner(dls, arch:str, loss_func=None, pretrained=True, cut=None, splitter=None,
                y_range=None, config=None, n_out=None, normalize=True, **kwargs):
    "Build a convnet style learner from `dls` and `arch` using the `timm` library"
    if config is None: config = {}
    if n_out is None: n_out = get_c(dls)
    assert n_out, "`n_out` is not defined, and could not be inferred from data, set `dls.c` or pass `n_out`"
    if y_range is None and 'y_range' in config: y_range = config.pop('y_range')
    model = create_timm_model(arch, n_out, default_split, pretrained, y_range=y_range, **config)
    learn = Learner(dls, model, loss_func=loss_func, splitter=default_split, **kwargs)
    if pretrained: learn.freeze()
    return learn


# AUTOGENERATED! DO NOT EDIT! File to edit: 04_lr_finder.ipynb (unless otherwise specified).

__all__ = []

# Cell
from fastai.vision.all import *

# Cell
@patch
def plot_lr_find(self:Recorder, suggestions= False,skip_end=5, lr_min=None, lr_steep=None):
    "Plot the result of an LR Finder test (won't work if you didn't do `learn.lr_find()` before)"
    lrs    = self.lrs    if skip_end==0 else self.lrs   [:-skip_end]
    losses = self.losses if skip_end==0 else self.losses[:-skip_end]

    if suggestions:
        lr_min_index = min(range(len(lrs)), key=lambda i: abs(lrs[i]-lr_min))
        lr_steep_index = min(range(len(lrs)), key=lambda i: abs(lrs[i]-lr_steep))

    fig, ax = plt.subplots(1,1)
    ax.plot(lrs, losses)
    if suggestions:
        ax.plot(lr_min,L(losses)[lr_min_index],'ro')
        ax.plot(lr_steep,L(losses)[lr_steep_index],'ro')
    ax.set_ylabel("Loss")
    ax.set_xlabel("Learning Rate")
    ax.set_xscale('log')

@patch
def lr_find(self:Learner, start_lr=1e-7, end_lr=10, num_it=100, stop_div=True, show_plot=True, suggestions=True):
    "Launch a mock training to find a good learning rate, return lr_min, lr_steep if `suggestions` is True"
    n_epoch = num_it//len(self.dls.train) + 1
    cb=LRFinder(start_lr=start_lr, end_lr=end_lr, num_it=num_it, stop_div=stop_div)
    with self.no_logging(): self.fit(n_epoch, cbs=cb)
    if suggestions:
        lrs,losses = tensor(self.recorder.lrs[num_it//10:-5]),tensor(self.recorder.losses[num_it//10:-5])
        if len(losses) == 0: return
        lr_min = lrs[losses.argmin()].item()
        grads = (losses[1:]-losses[:-1]) / (lrs[1:].log()-lrs[:-1].log())
        lr_steep = lrs[grads.argmin()].item()
        if show_plot: self.recorder.plot_lr_find(suggestions=True, lr_min =lr_min/10., lr_steep=lr_steep)
    else:
        if show_plot: self.recorder.plot_lr_find()

    if suggestions:
        return SuggestedLRs(lr_min/10.,lr_steep)


# AUTOGENERATED! DO NOT EDIT! File to edit: 00_utils.ipynb (unless otherwise specified).

__all__ = ['get_version', 'state_versions']

# Cell
import pkg_resources
def get_version(lib:str):
    "Returns version of `lib`"
    return pkg_resources.get_distribution(lib).version

# Cell
from IPython.display import Markdown
def state_versions(libs:list=[]):
    "State all the versions currently installed from `libs` in Markdown"
    cell = f"""
---
This article is also a Jupyter Notebook available to be run from the top down. There
will be code snippets that you can then run in any environment.

"""
    cell += 'Below are the versions of '
    for i, lib in enumerate(libs):
        if len(libs) == 1: cell += f'`{str(lib)}`'
        elif i < len(libs)-1:
            cell += ''.join(f'`{str(lib)}`')
            if len(libs) > 2: cell += ', '
            else: cell += ' '
        elif len(libs) > 1: cell += ''.join(f'and `{str(lib)}`')
    cell += ' currently running at the time of writing this:\n'
    cell += ''.join([
        f'* `{str(lib)}`: {get_version(str(lib))} \n'
    for lib in libs])
    cell += '---'
    return Markdown(cell)


# AUTOGENERATED! DO NOT EDIT! File to edit: 03_tab.stats.ipynb (unless otherwise specified).

__all__ = ['Categorify', 'setups', 'FillMissing']

# Cell
from fastai.tabular.all import *

# Cell
from fastai.tabular.core import _apply_cats

# Cell
class Categorify(TabularProc):
    "Transform the categorical variables to something similar to `pd.Categorical`"
    order = 1
    def __init__(self, classes=None):
        if classes is None: classes = defaultdict(L)
        store_attr()
        super().__init__()
    def setups(self, to):
        for n in to.cat_names:
            if n not in self.classes or is_categorical_dtype(to[n]):
                self.classes[n] = CategoryMap(to.iloc[:,n].items, add_na=n)

    def encodes(self, to): to.transform(to.cat_names, partial(_apply_cats, self.classes, 1))
    def decodes(self, to): to.transform(to.cat_names, partial(_decode_cats, self.classes))
    def __getitem__(self,k): return self.classes[k]

# Cell
@patch
def __init__(x:Normalize, mean=None, std=None, axes=(0,2,3), means=None, stds=None): store_attr()

# Cell
@Normalize
def setups(self, to:Tabular):
    if self.means is None:
        self.means = dict(getattr(to, 'train', to).conts.mean())
        self.stds  = dict(getattr(to, 'train', to).conts.std(ddof=0)+1e-7)
    else:
        for name in getattr(to, 'train', to).conts.columns:
            if name not in self.means.keys():
                self.means[name] = getattr(to, 'train', to)[name].mean()
            if name not in self.stds.keys():
                self.stds[name] = getattr(to, 'train', to)[name].std(ddof=0)+1e-7

    return self(to)

# Cell
class FillMissing(TabularProc):
    "Fill the missing values in continuous columns."
    def __init__(self, fill_strategy=FillStrategy.median, add_col=True, fill_vals=None, na_dict=None):
        if fill_vals is None: fill_vals = defaultdict(int)
        store_attr()

    def setups(self, to):
        missing = pd.isnull(to.conts).any()
        if self.na_dict is None:
            store_attr(na_dict={n:self.fill_strategy(to[n], self.fill_vals[n])
                                for n in missing[missing].keys()})
        else:
            diff_cols = list(set(to.cont_names) - set(self.na_dict.keys()))
            for key in diff_cols:
                self.na_dict[key] = self.fill_strategy(to[key], self.fill_vals[key])
        self.fill_strategy = self.fill_strategy.__name__

    def encodes(self, to):
        missing = pd.isnull(to.conts)
        for n in missing.any()[missing.any()].keys():
            assert n in self.na_dict, f"nan values in `{n}` but not in setup training set"
        for n in self.na_dict.keys():
            to[n].fillna(self.na_dict[n], inplace=True)
            if self.add_col:
                to.loc[:,n+'_na'] = missing[n]
                if n+'_na' not in to.cat_names: to.cat_names.append(n+'_na')